// Generated 11-Mar-2019 using Moditect maven plugin
module com.fasterxml.jackson.jr.ob {
    requires transitive com.fasterxml.jackson.core;

    exports com.fasterxml.jackson.jr.ob;
    exports com.fasterxml.jackson.jr.ob.api;
    exports com.fasterxml.jackson.jr.ob.comp;
    // 11-Mar-2019, tatu: Need to refactor and then close this one
    exports com.fasterxml.jackson.jr.ob.impl;
    exports com.fasterxml.jackson.jr.type;

}
