// Generated using Moditect maven plugin
module com.fasterxml.jackson.jr.stree {
    requires transitive com.fasterxml.jackson.core;
    requires com.fasterxml.jackson.jr.ob;
    requires com.fasterxml.jackson.jr.ob.api;

    exports com.fasterxml.jackson.jr.stree;
    exports com.fasterxml.jackson.jr.stree.util;
}
